package org.anarres.iengine.preprocess;

import java.io.IOException;
import java.util.Stack;
import static org.anarres.iengine.preprocess.Token.*;

%%

%class Lexer
%unicode
%line
%column
%final
%buffer 1024
%type Token
%debug

%yylexthrow{
LexerException
%yylexthrow}

%{
	private StringBuilder	buf = new StringBuilder();
	private Stack<Integer>	states = new Stack<Integer>();

	private String string() {
		String	str = buf.toString();
		buf.setLength(0);
		return str;
	}

	private void yy_push_state(int state) {
		states.push(Integer.valueOf(yystate()));
		yybegin(state);
	}

	private void yy_pop_state() {
		yybegin(states.pop().intValue());
	}

	private void yy_switch_state(int state) {
		yybegin(state);
	}

	private void yy_switch_state(int state, int after) {
		states.push(Integer.valueOf(after));
		yybegin(state);
	}

	private Token token(int type) {
		return new Token(type, yyline, yycolumn);
	}

	private Token token(int type, Object value) {
		return new Token(type, yyline, yycolumn, value);
	}

	private void exception(String msg) throws LexerException {
		throw new LexerException(yyline, yycolumn, msg);
	}

	public void expect(int type)
						throws IOException,
								LexerException {
		Token	token = yylex();
		if (token.getType() != type)
			exception("Expected token type " + type +
							" but got " + token);
	}
%}

/* main character classes */
LineTerminator = \r|\n|\r\n|\u2028|\u2029|\u000B|\u000C|\u0085
InputCharacter = [^\r\n]

WhiteSpace = {LineTerminator} | [ \t\f]

/* comments */
Comment = {TraditionalComment} | {EndOfLineComment}

TraditionalComment = "/*" ~"*/"
EndOfLineComment = "//" {InputCharacter}* {LineTerminator}?

/* identifiers */
Identifier = [:jletter:][:jletterdigit:]*

/* integer literals */
/*
DecIntegerLiteral = 0 | [1-9][0-9]*
DecLongLiteral    = {DecIntegerLiteral} [lL]

HexIntegerLiteral = 0 [xX] 0* {HexDigit} {1,8}
HexLongLiteral    = 0 [xX] 0* {HexDigit} {1,16} [lL]
HexDigit          = [0-9a-fA-F]

OctIntegerLiteral = 0+ [1-3]? {OctDigit} {1,15}
OctLongLiteral    = 0+ 1? {OctDigit} {1,21} [lL]
*/
OctDigit          = [0-7]

/* string and character literals */
StringCharacter = [^\r\n\"\\]
SingleCharacter = [^\r\n\'\\]

%s STRING, CHAR
%s PR, PR_INCLUDE, PR_INCLUDE_ANGLE,
%s PR_DEFINE, PR_DEFINE_SPACE, PR_DEFINE_ARGS, PR_DEFINE_VALUE
%s PR_IDENT, PR_EXP, PR_IGNORE

%%

<YYINITIAL> {
	{Comment}                { return token(COMMENT, yytext()); }
	{WhiteSpace}             { return token(WHITESPACE, yytext()); }
	{Identifier}             { /* XXX lookup */ return token(IDENTIFIER, yytext()); }
	^[:space:]*#             { yy_switch_state(PR); }
}

<PR> {
	include { yy_switch_state(PR_INCLUDE); return token(INCLUDE); }
	define  { yy_switch_state(PR_DEFINE);  return token(DEFINE); }
	undef   { yy_switch_state(PR_IDENT);   return token(UNDEF); }
	if      { yy_switch_state(PR_EXP);     return token(IF); }
	ifdef   { yy_switch_state(PR_IDENT);   return token(IFDEF); }
	ifndef  { yy_switch_state(PR_IDENT);   return token(IFNDEF); }
	else    { yy_switch_state(PR_IGNORE);  return token(ELSE); }
	endif   { yy_switch_state(PR_IGNORE);  return token(ENDIF); }
}

<PR_EXP> {
	\\{LineTerminator}   { }
	{LineTerminator}     { yy_switch_state(YYINITIAL); }
	.                    { /* XXX */ }
}

<PR_DEFINE> {
	\\{LineTerminator}   { }
	{WhiteSpace}         { }
	{Identifier}         {
		yy_switch_state(PR_DEFINE_SPACE);
		return token(IDENTIFIER, yytext());
	}
}

<PR_DEFINE_SPACE> {
	\\{LineTerminator}   { }
	{LineTerminator}     { yy_switch_state(YYINITIAL); }
	{WhiteSpace}         { yy_switch_state(PR_DEFINE_VALUE); }
	\(                   { yy_switch_state(PR_DEFINE_ARGS); }
}

<PR_DEFINE_ARGS> {
	\\{LineTerminator}   { }
	{LineTerminator}     { yy_switch_state(YYINITIAL); }
	{WhiteSpace}         { }
	{Identifier}         { return token(IDENTIFIER, yytext()); }
	,                    { return token(COMMA); }
	\)                   { yy_switch_state(PR_DEFINE_VALUE); }
}

<PR_DEFINE_VALUE> {
	\\{LineTerminator}   { buf.append(' '); }
	{LineTerminator}     { yy_switch_state(YYINITIAL); }
	{Identifier}         { /* XXX lookup */ return token(IDENTIFIER, yytext()); }
	##                   { return token(PASTE); }
}

<PR_IGNORE> {
	\\{LineTerminator}   { }
	{LineTerminator}     { yy_switch_state(YYINITIAL); }
}

<PR_INCLUDE> {
	\\{LineTerminator}   { }
	{WhiteSpace}         { }
	\<                   {
		yy_switch_state(PR_INCLUDE_ANGLE);
		return token(LANGLE);
	}
	\"                   {
		yy_switch_state(STRING, PR_IGNORE);
		return token(DQUOTE);
	}
}

<PR_IDENT> {
	{LineTerminator}     { yy_switch_state(YYINITIAL); }
	{WhiteSpace}         { }
	{Identifier}         { return token(IDENTIFIER, yytext()); }
}

<PR_INCLUDE_ANGLE> {
	[^\r\n\"\\>]+        { buf.append( yytext() ); }

	\>                   {
		yy_switch_state(PR_IGNORE);
		return token(STRING, string());
	}
}

<STRING> {
	{StringCharacter}+   { buf.append( yytext() ); }

	\"                   {
		yy_pop_state();
		return token(STRING, string());
	}
}

<STRING,PR_INCLUDE_ANGLE> {
  
	/* escape sequences */
	"\\b"                { buf.append( '\b' ); }
	"\\t"                { buf.append( '\t' ); }
	"\\n"                { buf.append( '\n' ); }
	"\\f"                { buf.append( '\f' ); }
	"\\r"                { buf.append( '\r' ); }
	"\\\""               { buf.append( '\"' ); }
	"\\'"                { buf.append( '\'' ); }
	"\\\\"               { buf.append( '\\' ); }
	\\[0-3]?{OctDigit}?{OctDigit}  {
		char val = (char)Integer.parseInt(yytext().substring(1), 8);
		buf.append( val );
	}

	/* error cases */
	\\.                            {
		exception("Illegal escape sequence \""+yytext()+"\"");
	}
	{LineTerminator}               {
		 exception("Unterminated string at end of line");
	}
}


.|\n         { exception("Illegal character \""+yytext()+ "\""); }
<<EOF>>      { return token(EOF); }
